"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_authorizer
hub.exec.boto3.client.apigatewayv2.delete_authorizer
hub.exec.boto3.client.apigatewayv2.get_apis
hub.exec.boto3.client.apigatewayv2.get_authorizer
hub.exec.boto3.client.apigatewayv2.get_authorizers
hub.exec.boto3.client.apigatewayv2.update_authorizer
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    api_id: str,
    authorizer_type: str,
    identity_source: List[str],
    resource_id: str = None,
    authorizer_credentials_arn: str = None,
    authorizer_payload_format_version: str = None,
    authorizer_result_ttl_in_seconds: int = None,
    authorizer_uri: str = None,
    enable_simple_responses: bool = None,
    jwt_configuration: make_dataclass(
        "JWTConfiguration",
        [
            ("Audience", List[str], field(default=None)),
            ("Issuer", str, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 authorized resource.

    Args:
        name(string): An Idem name of the resource. This is also used as the name of the Authorizer during resource creation.
        api_id(string): The API resource identifier in Amazon Web Services.
        authorizer_type(string): The authorizer type. Specify REQUEST for a Lambda function using incoming request parameters.
            Specify JWT to use JSON Web Tokens (supported only for HTTP APIs).
        resource_id(string, optional): The authorizer resource identifier in Amazon Web Services.
        authorizer_credentials_arn(str, optional): Specifies the required credentials as an IAM role for API Gateway to invoke the authorizer. To
            specify an IAM role for API Gateway to assume, use the role's Amazon Resource Name (ARN). To use
            resource-based permissions on the Lambda function, don't specify this parameter. Supported only
            for REQUEST authorizers.
        authorizer_payload_format_version(str, optional): Specifies the format of the payload sent to an HTTP API Lambda authorizer. Required for HTTP API
            Lambda authorizers. Supported values are 1.0 and 2.0. To learn more, see Working with AWS Lambda
            authorizers for HTTP APIs.
        authorizer_result_ttl_in_seconds(int, optional): The time to live (TTL) for cached authorizer results, in seconds. If it equals 0, authorization
            caching is disabled. If it is greater than 0, API Gateway caches authorizer responses. The
            maximum value is 3600, or 1 hour. Supported only for HTTP API Lambda authorizers.
        authorizer_type(str): The authorizer type. Specify REQUEST for a Lambda function using incoming request parameters.
            Specify JWT to use JSON Web Tokens (supported only for HTTP APIs).
        authorizer_uri(str, optional): The authorizer's Uniform Resource Identifier (URI). For REQUEST authorizers, this must be a
            well-formed Lambda function URI, for example, arn:aws:apigateway:us-
            west-2:lambda:path/2015-03-31/functions/arn:aws:lambda:us-
            west-2:{account_id}:function:{lambda_function_name}/invocations. In general, the URI has this
            form: arn:aws:apigateway:{region}:lambda:path/{service_api}                , where {region} is
            the same as the region hosting the Lambda function, path indicates that the remaining substring
            in the URI should be treated as the path to the resource, including the initial /. For Lambda
            functions, this is usually of the form /2015-03-31/functions/[FunctionARN]/invocations.
            Supported only for REQUEST authorizers.
        enable_simple_responses(bool, optional): Specifies whether a Lambda authorizer returns a response in a simple format. By default, a
            Lambda authorizer must return an IAM policy. If enabled, the Lambda authorizer can return a
            boolean value instead of an IAM policy. Supported only for HTTP APIs. To learn more, see Working
            with AWS Lambda authorizers for HTTP APIs.
        identity_source(List): The identity source for which authorization is requested. For a REQUEST authorizer, this is
            optional. The value is a set of one or more mapping expressions of the specified request
            parameters. The identity source can be headers, query string parameters, stage variables, and
            context parameters. For example, if an Auth header and a Name query string parameter are defined
            as identity sources, this value is route.request.header.Auth, route.request.querystring.Name for
            WebSocket APIs. For HTTP APIs, use selection expressions prefixed with $, for example,
            $request.header.Auth, $request.querystring.Name. These parameters are used to perform runtime
            validation for Lambda-based authorizers by verifying all of the identity-related request
            parameters are present in the request, not null, and non-empty. Only when this is true does the
            authorizer invoke the authorizer Lambda function. Otherwise, it returns a 401 Unauthorized
            response without calling the Lambda function. For HTTP APIs, identity sources are also used as
            the cache key when caching is enabled. To learn more, see Working with AWS Lambda authorizers
            for HTTP APIs. For JWT, a single entry that specifies where to extract the JSON Web Token (JWT)
            from inbound requests. Currently only header-based and query parameter-based selections are
            supported, for example $request.header.Authorization.
        jwt_configuration(Dict[str, Any], optional): Represents the configuration of a JWT authorizer. Required for the JWT authorizer type.
            Supported only for HTTP APIs. Defaults to None.
            * Audience (List[str], optional): A list of the intended recipients of the JWT. A valid JWT must provide an aud that matches at
                least one entry in this list. See RFC 7519. Supported only for HTTP APIs.
            * Issuer (str, optional): The base domain of the identity provider that issues JSON Web Tokens. For example, an Amazon
                Cognito user pool has the following format: https://cognito-idp.{region}.amazonaws.com/{userPoolId}.
                Required for the JWT authorizer type. Supported only for HTTP APIs.

    Request Syntax:
        [idem_test_aws_apigatewayv2_authorizer]:
          aws.apigatewayv2.authorizer.present:
            - name: 'string'
            - api_id: 'string'
            - authorizer_type: 'REQUEST'|'JWT'
            - identity_source: ['string']
            - authorizer_credentials_arn: 'string'
            - authorizer_payload_format_version: 'string'
            - authorizer_result_ttl_in_seconds: 123
            - authorizer_uri: 'string'
            - enable_simple_responses: True|False
            - jwt_configuration: {
                'Audience': ['string'],
                'Issuer': 'string'
              }

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_authorizer:
              aws.apigatewayv2.authorizer.present:
                - name: value
                - api_id: value
                - authorizer_type: value
                - identity_source: [value]
                - authorizer_credentials_arn: value
                - authorizer_payload_format_version: value
                - authorizer_result_ttl_in_seconds: value
                - authorizer_uri: value
                - enable_simple_responses: True|False
                - jwt_configuration: {
                    'Audience': [value],
                    'Issuer': value
                  }
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.boto3.client.apigatewayv2.get_authorizer(
            ctx, ApiId=api_id, AuthorizerId=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.authorizer", name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.apigatewayv2.authorizer.convert_raw_authorizer_to_present(
            api_id=api_id, raw_resource=before_ret.get("ret")
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "AuthorizerCredentialsArn": authorizer_credentials_arn,
            "AuthorizerPayloadFormatVersion": authorizer_payload_format_version,
            "AuthorizerResultTtlInSeconds": authorizer_result_ttl_in_seconds,
            "AuthorizerType": authorizer_type,
            "AuthorizerUri": authorizer_uri,
            "EnableSimpleResponses": enable_simple_responses,
            "IdentitySource": identity_source,
            "JwtConfiguration": jwt_configuration,
        }

        update_authorizer_ret = await hub.exec.aws.apigatewayv2.authorizer.update(
            ctx,
            api_id=api_id,
            resource_id=resource_id,
            raw_resource=before_ret.get("ret"),
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_authorizer_ret["comment"]
        if not update_authorizer_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_authorizer_ret["ret"])
        if update_authorizer_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_authorizer_ret["ret"])

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "api_id": api_id,
                    "authorizer_credentials_arn": authorizer_credentials_arn,
                    "authorizer_payload_format_version": authorizer_payload_format_version,
                    "authorizer_result_ttl_in_seconds": authorizer_result_ttl_in_seconds,
                    "authorizer_type": authorizer_type,
                    "authorizer_uri": authorizer_uri,
                    "enable_simple_responses": enable_simple_responses,
                    "identity_source": identity_source,
                    "jwt_configuration": jwt_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.authorizer", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_authorizer(
            ctx,
            ApiId=api_id,
            AuthorizerCredentialsArn=authorizer_credentials_arn,
            AuthorizerPayloadFormatVersion=authorizer_payload_format_version,
            AuthorizerResultTtlInSeconds=authorizer_result_ttl_in_seconds,
            AuthorizerType=authorizer_type,
            AuthorizerUri=authorizer_uri,
            EnableSimpleResponses=enable_simple_responses,
            IdentitySource=identity_source,
            JwtConfiguration=jwt_configuration,
            Name=name,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.authorizer", name=name
        )
        resource_id = create_ret["ret"]["AuthorizerId"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.apigatewayv2.get_authorizer(
            ctx, ApiId=api_id, AuthorizerId=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result[
            "new_state"
        ] = hub.tool.aws.apigatewayv2.authorizer.convert_raw_authorizer_to_present(
            api_id=api_id, raw_resource=after_ret.get("ret")
        )

    return result


async def absent(
    hub, ctx, name: str, api_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 authorizer resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string, optional): The Authorizer resource identifier in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_authorizer:
              aws.apigatewayv2.authorizer.absent:
                - name: value
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.authorizer", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigatewayv2.get_authorizer(
        ctx, ApiId=api_id, AuthorizerId=resource_id
    )
    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigatewayv2.authorizer", name=name
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.apigatewayv2.authorizer.convert_raw_authorizer_to_present(
        api_id=api_id, raw_resource=before_ret["ret"]
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.authorizer", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_authorizer(
        ctx, ApiId=api_id, AuthorizerId=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigatewayv2.authorizer", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 authorized resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.authorizer
    """

    result = {}

    get_apis_ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)
    if not get_apis_ret["result"]:
        hub.log.debug(f"Could not get apis {get_apis_ret['comment']}")
        return result

    for api in get_apis_ret["ret"]["Items"]:
        api_id = api.get("ApiId")

        get_authorizers_ret = await hub.exec.boto3.client.apigatewayv2.get_authorizers(
            ctx, ApiId=api_id
        )
        if not get_authorizers_ret["result"]:
            hub.log.debug(
                f"Could not get authorizers for ApiId '{api_id}': "
                f"{get_authorizers_ret['comment']}. Describe will skip this api and continue."
            )
            continue

        for authorizer in get_authorizers_ret["ret"]["Items"]:
            resource_translated = (
                hub.tool.aws.apigatewayv2.authorizer.convert_raw_authorizer_to_present(
                    api_id=api_id,
                    raw_resource=authorizer,
                )
            )

            result[resource_translated["name"]] = {
                "aws.apigatewayv2.authorizer.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }

    return result


async def search(hub, ctx, name, api_id: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Use an un-managed API Gateway v2 authorizer resource as a data-source.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Authorizer resource identifier in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_authorizer]:
          aws.apigatewayv2.authorizer.search:
            - api_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_authorizer:
              aws.apigatewayv2.authorizer.search:
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    search_ret = await hub.exec.boto3.client.apigatewayv2.get_authorizer(
        ctx, ApiId=api_id, AuthorizerId=resource_id
    )
    if not search_ret["result"]:
        result["result"] = False
        if "NotFoundException" not in str(search_ret["comment"][0]):
            result["comment"] = search_ret["result"]
        else:
            result["comment"] = (
                f"Unable to find aws.apigatewayv2.authorizer resource with resource id '{resource_id}'",
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.apigatewayv2.authorizer.convert_raw_authorizer_to_present(
        api_id=api_id, raw_resource=search_ret["ret"]
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.autoscaling.create_launch_configuration
hub.exec.boto3.client.autoscaling.delete_launch_configuration
hub.exec.boto3.client.autoscaling.describe_launch_configurations
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    image_id: str = None,
    key_name: str = None,
    security_groups: List[str] = None,
    classic_link_vpc_id: str = None,
    classic_link_vpc_security_groups: List[str] = None,
    user_data: str = None,
    instance_id: str = None,
    instance_type: str = None,
    kernel_id: str = None,
    ramdisk_id: str = None,
    block_device_mappings: List[
        make_dataclass(
            "BlockDeviceMapping",
            [
                ("DeviceName", str),
                ("VirtualName", str, field(default=None)),
                (
                    "Ebs",
                    make_dataclass(
                        "Ebs",
                        [
                            ("SnapshotId", str, field(default=None)),
                            ("VolumeSize", int, field(default=None)),
                            ("VolumeType", str, field(default=None)),
                            ("DeleteOnTermination", bool, field(default=None)),
                            ("Iops", int, field(default=None)),
                            ("Encrypted", bool, field(default=None)),
                            ("Throughput", int, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("NoDevice", bool, field(default=None)),
            ],
        )
    ] = None,
    instance_monitoring: make_dataclass(
        "InstanceMonitoring", [("Enabled", bool, field(default=None))]
    ) = None,
    spot_price: str = None,
    iam_instance_profile: str = None,
    ebs_optimized: bool = None,
    associate_public_ip_address: bool = None,
    placement_tenancy: str = None,
    metadata_options: make_dataclass(
        "InstanceMetadataOptions",
        [
            ("HttpTokens", str, field(default=None)),
            ("HttpPutResponseHopLimit", int, field(default=None)),
            ("HttpEndpoint", str, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a launch configuration. If you exceed your maximum limit of launch configurations, the call fails. To
    query this limit, call the DescribeAccountLimits API. For information about updating this limit, see Amazon EC2
    Auto Scaling service quotas in the Amazon EC2 Auto Scaling User Guide. For more information, see Launch
    configurations in the Amazon EC2 Auto Scaling User Guide.

    Args:
        name(Text): The name of the launch configuration. This name must be unique per Region per account.
        resource_id(Text, optional): AWS Launch Configuration name
        image_id(Text, optional): The ID of the Amazon Machine Image (AMI) that was assigned during registration. For more
            information, see Finding an AMI in the Amazon EC2 User Guide for Linux Instances. If you do not
            specify InstanceId, you must specify ImageId. Defaults to None.
        key_name(Text, optional): The name of the key pair. For more information, see Amazon EC2 Key Pairs in the Amazon EC2 User
            Guide for Linux Instances. Defaults to None.
        security_groups(List, optional): A list that contains the security groups to assign to the instances in the Auto Scaling group.
            [EC2-VPC] Specify the security group IDs. For more information, see Security Groups for Your VPC
            in the Amazon Virtual Private Cloud User Guide. [EC2-Classic] Specify either the security group
            names or the security group IDs. For more information, see Amazon EC2 Security Groups in the
            Amazon EC2 User Guide for Linux Instances. Defaults to None.
        classic_link_vpc_id(Text, optional): The ID of a ClassicLink-enabled VPC to link your EC2-Classic instances to. For more information,
            see ClassicLink in the Amazon EC2 User Guide for Linux Instances and Linking EC2-Classic
            instances to a VPC in the Amazon EC2 Auto Scaling User Guide. This parameter can only be used if
            you are launching EC2-Classic instances. Defaults to None.
        classic_link_vpc_security_groups(List, optional): The IDs of one or more security groups for the specified ClassicLink-enabled VPC. For more
            information, see ClassicLink in the Amazon EC2 User Guide for Linux Instances and Linking
            EC2-Classic instances to a VPC in the Amazon EC2 Auto Scaling User Guide. If you specify the
            ClassicLinkVPCId parameter, you must specify this parameter. Defaults to None.
        user_data(Text, optional): The user data to make available to the launched EC2 instances. For more information, see
            Instance metadata and user data (Linux) and Instance metadata and user data (Windows). If you
            are using a command line tool, base64-encoding is performed for you, and you can load the text
            from a file. Otherwise, you must provide base64-encoded text. User data is limited to 16 KB. Defaults to None.
        instance_id(Text, optional): The ID of the instance to use to create the launch configuration. The new launch configuration
            derives attributes from the instance, except for the block device mapping. To create a launch
            configuration with a block device mapping or override any other instance attributes, specify
            them as part of the same request. For more information, see Creating a launch configuration
            using an EC2 instance in the Amazon EC2 Auto Scaling User Guide. If you do not specify
            InstanceId, you must specify both ImageId and InstanceType. Defaults to None.
        instance_type(Text, optional): Specifies the instance type of the EC2 instance. For information about available instance types,
            see Available Instance Types in the Amazon EC2 User Guide for Linux Instances. If you do not
            specify InstanceId, you must specify InstanceType. Defaults to None.
        kernel_id(Text, optional): The ID of the kernel associated with the AMI. Defaults to None.
        ramdisk_id(Text, optional): The ID of the RAM disk to select. Defaults to None.
        block_device_mappings(List, optional): A block device mapping, which specifies the block devices for the instance. You can specify
            virtual devices and EBS volumes. For more information, see Block Device Mapping in the Amazon
            EC2 User Guide for Linux Instances. Defaults to None.
            * (VirtualName, optional): The name of the virtual device (for example, ephemeral0). You can specify either VirtualName or
            Ebs, but not both.
            * (DeviceName): The device name exposed to the EC2 instance (for example, /dev/sdh or xvdh). For more
            information, see Device Naming on Linux Instances in the Amazon EC2 User Guide for Linux
            Instances.
            * (Ebs, optional): Parameters used to automatically set up EBS volumes when an instance is launched. You can
            specify either VirtualName or Ebs, but not both.
            * (NoDevice, optional): Setting this value to true suppresses the specified device included in the block device mapping
            of the AMI. If NoDevice is true for the root device, instances might fail the EC2 health check.
            In that case, Amazon EC2 Auto Scaling launches replacement instances. If you specify NoDevice,
            you cannot specify Ebs.
        instance_monitoring(Dict, optional): Controls whether instances in this group are launched with detailed (true) or basic (false)
            monitoring. The default value is true (enabled).  When detailed monitoring is enabled, Amazon
            CloudWatch generates metrics every minute and your account is charged a fee. When you disable
            detailed monitoring, CloudWatch generates metrics every 5 minutes. For more information, see
            Configure Monitoring for Auto Scaling Instances in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
            * (Enabled, optional): If true, detailed monitoring is enabled. Otherwise, basic monitoring is enabled.
        spot_price(Text, optional): The maximum hourly price to be paid for any Spot Instance launched to fulfill the request. Spot
            Instances are launched when the price you specify exceeds the current Spot price. For more
            information, see Requesting Spot Instances in the Amazon EC2 Auto Scaling User Guide.  When you
            change your maximum price by creating a new launch configuration, running instances will
            continue to run as long as the maximum price for those running instances is higher than the
            current Spot price. Defaults to None.
        iam_instance_profile(Text, optional): The name or the Amazon Resource Name (ARN) of the instance profile associated with the IAM role
            for the instance. The instance profile contains the IAM role. For more information, see IAM role
            for applications that run on Amazon EC2 instances in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        ebs_optimized(bool, optional): Specifies whether the launch configuration is optimized for EBS I/O (true) or not (false). The
            optimization provides dedicated throughput to Amazon EBS and an optimized configuration stack to
            provide optimal I/O performance. This optimization is not available with all instance types.
            Additional fees are incurred when you enable EBS optimization for an instance type that is not
            EBS-optimized by default. For more information, see Amazon EBS-optimized instances in the Amazon
            EC2 User Guide for Linux Instances. The default value is false. Defaults to None.
        associate_public_ip_address(bool, optional): For Auto Scaling groups that are running in a virtual private cloud (VPC), specifies whether to
            assign a public IP address to the group's instances. If you specify true, each instance in the
            Auto Scaling group receives a unique public IP address. For more information, see Launching Auto
            Scaling instances in a VPC in the Amazon EC2 Auto Scaling User Guide. If you specify this
            parameter, you must specify at least one subnet for VPCZoneIdentifier when you create your
            group.  If the instance is launched into a default subnet, the default is to assign a public IP
            address, unless you disabled the option to assign a public IP address on the subnet. If the
            instance is launched into a nondefault subnet, the default is not to assign a public IP address,
            unless you enabled the option to assign a public IP address on the subnet. Defaults to None.
        placement_tenancy(Text, optional): The tenancy of the instance. An instance with dedicated tenancy runs on isolated, single-tenant
            hardware and can only be launched into a VPC. To launch dedicated instances into a shared
            tenancy VPC (a VPC with the instance placement tenancy attribute set to default), you must set
            the value of this parameter to dedicated. If you specify PlacementTenancy, you must specify at
            least one subnet for VPCZoneIdentifier when you create your group. For more information, see
            Configuring instance tenancy with Amazon EC2 Auto Scaling in the Amazon EC2 Auto Scaling User
            Guide. Valid Values: default | dedicated. Defaults to None.
        metadata_options(Dict, optional): The metadata options for the instances. For more information, see Configuring the Instance
            Metadata Options in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
            * (HttpTokens, optional): The state of token usage for your instance metadata requests. If the parameter is not specified
                in the request, the default state is optional. If the state is optional, you can choose to
                retrieve instance metadata with or without a signed token header on your request. If you
                retrieve the IAM role credentials without a token, the version 1.0 role credentials are
                returned. If you retrieve the IAM role credentials using a valid signed token, the version 2.0
                role credentials are returned. If the state is required, you must send a signed token header
                with any instance metadata retrieval requests. In this state, retrieving the IAM role
                credentials always returns the version 2.0 credentials; the version 1.0 credentials are not
                available.
            * (HttpPutResponseHopLimit, optional): The desired HTTP PUT response hop limit for instance metadata requests. The larger the number,
                the further instance metadata requests can travel. Default: 1
            * (HttpEndpoint, optional): This parameter enables or disables the HTTP metadata endpoint on your instances. If the
                parameter is not specified, the default state is enabled.  If you specify a value of disabled,
                you will not be able to access your instance metadata.

    Request Syntax:
       [launch_configuration-resource-id]:
              aws.autoscaling.launch_configuration.present:
              - name: 'string'
              - resource_id: 'string'
              - image_id: 'string'
              - key_name: 'string'
              - security_groups: List
              - classic_link_vpc_id: 'string'
              - classic_link_vpc_security_groups: List
              - user_data: 'string'
              - instance_type: 'string'
              - kernel_id: aki-02b79b47
              - ramdisk_id: 'string'
              - block_device_mappings: List
              - instance_monitoring: Dict
              - spot_price: 'string'
              - iam_instance_profile: 'string'
              - ebs_optimized: true/false
              - metadata_options: Dict
                          - associate_public_ip_address: true/false

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-launch_configuration:
              aws.autoscaling.launch_configuration.present:
              - name: idem-test-launch_configuration
              - resource_id: idem-test-launch_configuration
              - image_id: ami-73949613
              - key_name: test-n
              - security_groups:
                - sg-00bbd7b8d31f6bcad
                - sg-03064434ef6546344
              - classic_link_vpc_id: vpc-123456
              - classic_link_vpc_security_groups:
                - sg-00bbd7b8d31f6abcd
                - sg-03064434ef6541111
              - user_data: /etc/eks/bootstrap.sh
              - instance_type: t2.micro
              - kernel_id: aki-02b79b47
              - ramdisk_id: ari-177e2d52
              - block_device_mappings:
                - DeviceName: /dev/sda1
                  Ebs:
                    DeleteOnTermination: false
                    Encrypted: false
                    SnapshotId: snap-0526fa8583b2ef05f
                    VolumeSize: 8
                    VolumeType: gp2
                - DeviceName: /dev/sda
                  Ebs:
                    DeleteOnTermination: false
                    Encrypted: false
                    SnapshotId: snap-05cfe4be1b16a2e0f
                    VolumeSize: 8
                    VolumeType: gp2
              - instance_monitoring:
                  Enabled: true
              - spot_price: '0.045'
              - iam_instance_profile: arn:aws:iam::537227425989:instance-profile/aws-elasticbeanstalk-ec2-role
              - ebs_optimized: true
              - associate_public_ip_address: true
              - metadata_options:
                  HttpEndpoint: enabled
                  HttpPutResponseHopLimit: 64
                  HttpTokens: optional

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    launch_configuration_created = False
    before = None
    desired_state = {
        "name": name,
        "resource_id": name,
        "image_id": image_id,
        "key_name": key_name,
        "security_groups": security_groups,
        "classic_link_vpc_id": classic_link_vpc_id,
        "classic_link_vpc_security_groups": classic_link_vpc_security_groups,
        "user_data": user_data,
        "instance_id": instance_id,
        "instance_type": instance_type,
        "kernel_id": kernel_id,
        "ramdisk_id": ramdisk_id,
        "block_device_mappings": block_device_mappings,
        "instance_monitoring": instance_monitoring,
        "spot_price": spot_price,
        "iam_instance_profile": iam_instance_profile,
        "ebs_optimized": ebs_optimized,
        "associate_public_ip_address": associate_public_ip_address,
        "placement_tenancy": placement_tenancy,
        "metadata_options": metadata_options,
    }

    launch_conf_args = {
        "LaunchConfigurationName": name,
        "ImageId": image_id,
        "KeyName": key_name,
        "SecurityGroups": security_groups,
        "ClassicLinkVPCId": classic_link_vpc_id,
        "ClassicLinkVPCSecurityGroups": classic_link_vpc_security_groups,
        "UserData": user_data,
        "InstanceId": instance_id,
        "InstanceType": instance_type,
        "KernelId": kernel_id,
        "RamdiskId": ramdisk_id,
        "BlockDeviceMappings": block_device_mappings,
        "InstanceMonitoring": instance_monitoring,
        "SpotPrice": spot_price,
        "IamInstanceProfile": iam_instance_profile,
        "EbsOptimized": ebs_optimized,
        "AssociatePublicIpAddress": associate_public_ip_address,
        "PlacementTenancy": placement_tenancy,
        "MetadataOptions": metadata_options,
    }

    if resource_id:
        before = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
            ctx, LaunchConfigurationNames=[resource_id]
        )

    # Launch Configurations cannot be updated after creation with the Amazon Web Service API.
    if before and before["ret"].get("LaunchConfigurations"):  # resource present
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=before["ret"]["LaunchConfigurations"][0],
            idem_resource_name=name,
        )
        result["comment"] = (
            f"aws.autoscaling.launch_configuration '{name}' already exists",
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"] = (
                f"Would create aws.autoscaling.launch_configuration '{name}'",
            )
            return result

        try:
            ret = await hub.exec.boto3.client.autoscaling.create_launch_configuration(
                ctx,
                **launch_conf_args,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            result["comment"] = (
                f"Created aws.autoscaling.launch_configuration '{name}'",
            )
            launch_configuration_created = True
            resource_id = name
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if launch_configuration_created:
            after = (
                await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
                    ctx, LaunchConfigurationNames=[resource_id]
                )
            )
            result[
                "new_state"
            ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
                raw_resource=after["ret"]["LaunchConfigurations"][0],
                idem_resource_name=name,
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified launch configuration. The launch configuration must not be attached to an Auto Scaling
    group. When this call completes, the launch configuration is no longer available for use.

    Args:
        name(Text): The name of the launch configuration.
        resource_id(Text, optional): AWS launch configuration name. Idem automatically considers this resource being absent if
         this field is not specified.

    Request Syntax:
       [launch_configuration-resource-id]:
              aws.autoscaling.launch_configuration.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-launch_configuration:
              aws.autoscaling.launch_configuration.absent:
                - name: idem-test-launch_configuration
                - resource_id:idem-test-launch_configuration
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.launch_configuration", name=name
        )
        return result

    before = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(
        ctx, LaunchConfigurationNames=[resource_id]
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"].get("LaunchConfigurations"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.launch_configuration", name=name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=before["ret"]["LaunchConfigurations"][0],
            idem_resource_name=name,
        )

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.autoscaling.launch_configuration", name=name
            )
            return result

        try:
            ret = await hub.exec.boto3.client.autoscaling.delete_launch_configuration(
                ctx, LaunchConfigurationName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.autoscaling.launch_configuration", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets information about the launch configurations in the account and Region.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.autoscaling.launch_configuration
    """

    result = {}
    ret = await hub.exec.boto3.client.autoscaling.describe_launch_configurations(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe launch_configuration {ret['comment']}")
        return {}

    for launch_configuration in ret["ret"].get("LaunchConfigurations"):
        resource_id = launch_configuration.get("LaunchConfigurationName")
        resource_translated = hub.tool.aws.autoscaling.conversion_utils.convert_raw_launch_configuration_to_present(
            raw_resource=launch_configuration, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.autoscaling.launch_configuration.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



import copy
from typing import Any
from typing import Dict


async def update_detector_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    tags_to_add = {}
    tags_to_remove = []
    tags_result = {}
    if old_tags:
        tags_result = copy.deepcopy(old_tags)
    if new_tags is not None:
        for key, value in new_tags.items():
            if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
                key not in old_tags
            ):
                tags_to_add[key] = value
    if old_tags:
        for key in old_tags:
            if key not in new_tags:
                tags_to_remove.append(key)
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.guardduty.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key, None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.guardduty.tag_resource(
                ctx, ResourceArn=resource_arn, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    tags_to_append = {**tags_result, **tags_to_add}
    result["ret"] = {"tags": tags_to_append}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result


async def update_detector(
    hub,
    ctx,
    before: Dict[str, Any],
    detector_id: str = None,
    finding_publishing_frequency: str = None,
    data_sources: Dict[str, Any] = None,
):
    result = dict(comment=(), result=True, ret=None)
    compare_dict_result = False
    update_payload = {}
    if (finding_publishing_frequency is not None) and before["ret"].get(
        "FindingPublishingFrequency"
    ) != finding_publishing_frequency:
        update_payload["FindingPublishingFrequency"] = finding_publishing_frequency
    data_describe = before["ret"]
    data_sources_rendered = hub.tool.aws.guardduty.detector_utils.render_data_sources(
        data_describe
    )
    if data_sources:
        compare_dict_result = hub.tool.aws.guardduty.detector_utils.compare_dicts(
            data_sources, data_sources_rendered
        )
    if not compare_dict_result and data_sources is not None:
        update_payload["DataSources"] = data_sources
    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.guardduty.update_detector(
                ctx=ctx, DetectorId=detector_id, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        if "FindingPublishingFrequency" in update_payload:
            result["ret"]["finding_publishing_frequency"] = update_payload[
                "FindingPublishingFrequency"
            ]
            result["comment"] = result["comment"] + (
                f"Update finding_publishing_frequency: {update_payload['FindingPublishingFrequency']}",
            )
        if "DataSources" in update_payload:
            result["ret"]["data_sources"] = update_payload["DataSources"]
            result["comment"] = result["comment"] + (
                f"Update data_sources: {update_payload['DataSources']}",
            )
    return result



from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import boto3.resources.factory
import boto3.session

__func_alias__ = {"exec_": "exec", "dict_": "dict", "id_": "id"}


def create(
    hub,
    ctx,
    service_name: str,
    resource_class: str,
    *args,
    **resource_kwargs: Dict[str, Any],
) -> boto3.resources.factory.ServiceResource:
    """
    Create a boto3 resource class

    :param hub:
    :param ctx:
    :param service_name: The name of the service to get from boto3
    :param resource_class: The name of the resource class to get from the service
    :param resource_kwargs: kwargs to pass to the resource class creation

    :return:
    """
    session: boto3.session.Session = hub.tool.boto3.session.get(
        # Resources are not thread safe, use a brand new session
        botocore_session=None
    )
    resource: boto3.resources.factory.ResourceFactory = session.resource(
        service_name=service_name,
        region_name=hub.tool.boto3.region.get(ctx),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )

    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in resource_kwargs.items() if v is not None}

    hub.log.debug(f"Getting resource {service_name}.{resource_class}")
    cls: boto3.resources.factory.ResourceHandler = getattr(resource, resource_class)
    if callable(cls):
        return cls(*args, **kwargs)
    else:
        return cls


class ResourceDict(dict):
    """
    Helper when outputting details of a resource class on the CLI
    """

    def __init__(self, resource: boto3.resources.factory.ServiceResource):
        super().__init__()
        self.resource = resource
        for i in self.resource.meta.identifiers:
            self[i] = getattr(self.resource, i)

    def __getattr__(self, item):
        return self.resource.__getattribute__(item)


def dict_(hub, resource: boto3.resources.factory.ServiceResource):
    # Helper on the hub for creating a resource dict
    return ResourceDict(resource)


async def exec_(
    hub,
    resource: boto3.resources.factory.ServiceResource,
    operation: str,
    *op_args,
    **op_kwargs: Dict[str, Any],
):
    """
    Call an operation on an existing resource

    :param hub:
    :param resource: The resource to call
    :param operation: The name of the operation to call on the resource
    :param op_args: Args to pass to the operation
    :param op_kwargs: Kwargs to pass to the operation
    :return:
    """
    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in op_kwargs.items() if v is not None}

    hub.log.debug(f"Getting resource {resource}.{operation}")
    op = getattr(resource, operation)
    if callable(op):
        # It's a function, run it asynchronously
        return await hub.pop.loop.wrap(op, *op_args, **kwargs)
    else:
        # It's an attribute
        return op


async def describe(hub, resource: boto3.resources.factory.ServiceResource):
    """
    Describe a boto3 resource

    :param hub:
    :param resource: The resource to call

    :return:
    """
    try:
        await hub.pop.loop.wrap(resource.load)
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"{resource} does not exist")
        return {}

    return resource.meta.data


async def id_(hub, ctx, service_name: str, collection: str, **kwargs) -> str:
    """
    Find the ID of a resource that matches the given kwargs in a collection
    :param hub:
    :param ctx:
    :param service_name: The name of the service to get from boto3
    :param collection: The name of the collection to search under the resource
    :param kwargs: The kwargs to pass to the filter function
    :return:
    """
    session: boto3.session.Session = hub.tool.boto3.session.get(
        # Resources are not thread safe, use a brand new session
        botocore_session=None
    )

    resource: boto3.resources.factory.ResourceFactory = session.resource(
        service_name=service_name,
        region_name=hub.tool.boto3.region.get(ctx),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )

    assert any(
        collection == c.name for c in resource.meta.resource_model.collections
    ), f"Service {service_name} has no collection '{collection}: {','.join(resource.meta.resource_model.collections)}'"
    collected = getattr(resource, collection)

    iterator = await hub.pop.loop.wrap(
        collected.filter,
        MaxResults=5,
        DryRun=ctx.get("test", False),
        Filters=[
            {"Name": k, "Values": v if isinstance(v, (List, Tuple)) else [v]}
            for k, v in kwargs.items()
        ],
    )

    try:
        first = next(iter(iterator))
        return first.id
    except StopIteration:
        return ""



import inspect
import json
from typing import Any
from typing import Dict
from typing import List


def are_lists_identical(hub, list1: List, list2: List) -> bool:
    """
    Compare two lists and logs the difference.
    :param list1: first list.
    :param list2: second list.
    :return: true if there is no difference between both lists.
    :raises exception if one of the list  is not of type list
    """
    if (list1 is None or len(list1) == 0) and (list2 is None or len(list2) == 0):
        return True
    if list1 is None or len(list1) == 0 or list2 is None or len(list2) == 0:
        return False

    for l in [list1, list2]:
        if not isinstance(l, List):
            raise TypeError(
                f"Expecting lists to compare. This is expected to be of type List: '{l}'"
            )

    diff = [i for i in list1 + list2 if i not in list1 or i not in list2]
    result = len(diff) == 0
    if not result:
        hub.log.debug(f"There are {len(diff)} differences:\n{diff[:5]}")
    return result


def standardise_json(hub, value: str or Dict, sort_keys: bool = True) -> str:
    # Format json string or dictionary
    if value is None or value is inspect._empty or not value:
        return None

    if isinstance(value, str) and len(value) > 0:
        json_dict = json.loads(value)
    elif isinstance(value, Dict):
        json_dict = value
    else:
        raise TypeError(
            f"Expecting string or dictionary. This value has the wrong type: '{value}'"
        )

    return json.dumps(
        _sorting(json_dict) if sort_keys else json_dict,
        separators=(", ", ": "),
        sort_keys=sort_keys,
    )


def is_json_identical(hub, struct1: str, struct2: str):
    return _sorting(json.loads(struct1)) == _sorting(json.loads(struct2))


def _sorting(obj):
    if isinstance(obj, dict):
        return {k: _sorting(v) for k, v in sorted(obj.items())}
    if isinstance(obj, list):
        if all(isinstance(v, str) for v in obj):
            return sorted(obj)
        else:
            return [_sorting(v) for i, v in sorted(enumerate(obj))]
    return obj


def compare_dicts(hub, source_dict: Dict[str, Any], target_dict: Dict[str, Any]):
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return true or false based on the comparison

    Returns:
        {True|False}

    """

    for key, value in source_dict.items():
        if key in target_dict:
            if isinstance(source_dict[key], dict):
                if not compare_dicts(hub, source_dict[key], target_dict[key]):
                    return False
            elif value != target_dict[key]:
                return False
        else:
            return False
    return True



"""
hub.exec.boto3.client.config.put_delivery_channel
hub.exec.boto3.client.config.delete_delete_channel
hub.exec.boto3.client.config.describe_delivery_channels

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    s3_bucket_name: str,
    resource_id: str = None,
    s3_key_prefix: str = None,
    s3_kms_key_arn: str = None,
    sns_topic_arn: str = None,
    config_snapshot_delivery_properties: make_dataclass(
        "ConfigSnapshotDeliveryProperties",
        [("delivery_frequency", str, field(default=None))],
    ) = None,
) -> Dict[str, Any]:
    """
    Add or update the configuration delivery channel object that delivers the configuration information to an Amazon S3 bucket
    and to an Amazon SNS topic. For more information about delivery channel, please see AWS Config services
    Args:
        name(Text): An Idem name of the delivery channel.
        s3_bucket_name (Text): The name of the Amazon S3 bucket to which Config delivers configuration snapshots and configuration history files
        resource_id (Text, Optional): The resource Id of the delivery channel.
        s3_key_prefix (Text, Optional): The prefix for the specified Amazon S3 bucket.
        s3_kms_key_arn (Text, Optional): The Amazon Resource Name (ARN) of the Key Management Service (KMS ) KMS key (KMS key) used to encrypt objects delivered by Config.
        sns_topic_arn (Text, Optional): The Amazon Resource Name (ARN) of the Amazon SNS topic to which Config sends notifications about configuration changes.
        config_snapshot_delivery_properties (Dict[str, Any], Optional):
            * delivery_frequency (str, optional): The frequency with which Config delivers configuration snapshots.

    Request syntax:
        [aws-config-delivery_channel]:
          aws.config.delivery_channel.present:
          - name: 'string'
          - resource_id: 'string'
          - s3_bucket_name 'string'
          - s3_key_prefix: 'string'
          - s3_kms_key_arn: 'string'
          - sns_topic_arn: 'string'
          - config_snapshot_delivery_properties: dict
            delivery_frequency: 'string'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls
            aws-config-delivery_channel:
              aws.config.delivery_channel.present:
                - name: 'delivery-channel'
                - resource_id: 'delivery-channel'
                - s3_bucket_name 'Test-S3'
                - s3_key_prefix: 'S3-prefix'
                - s3_kms_key_arn: 'Test-Kms-Key'
                - sns_topic_arn: 'Test-Topic'
                - config_snapshot_delivery_properties:
                  delivery_frequency: 'One_Hour'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    update_payload = {}
    update_config_snapshot_delivery_properties = None
    resource_id = resource_id if resource_id else name
    before = await hub.exec.boto3.client.config.describe_delivery_channels(
        ctx, DeliveryChannelNames=[resource_id]
    )
    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=before["ret"]["DeliveryChannels"][0],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])
        # updating config_snapshot_delivery_properties
        if config_snapshot_delivery_properties is not None:
            update_config_snapshot_delivery_properties = (
                config_snapshot_delivery_properties
            )
        update_payload = await hub.exec.aws.config.delivery_channel.get_updated_payload_delivery_channel(
            ctx,
            resource_id=resource_id,
            before=before["ret"]["DeliveryChannels"][0],
            s3_bucket_name=s3_bucket_name,
            s3_key_prefix=s3_key_prefix,
            s3_kms_key_arn=s3_kms_key_arn,
            sns_topic_arn=sns_topic_arn,
            config_snapshot_delivery_properties=update_config_snapshot_delivery_properties,
        )

        if update_payload["ret"]:
            resource_updated = True
            if not ctx.get("test", False):
                update_ret = await hub.exec.boto3.client.config.put_delivery_channel(
                    ctx=ctx, DeliveryChannel=update_payload["ret"]
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False
                    return result

                result["comment"] = (f"Updated aws.config.delivery_channel '{name}'",)
                result["result"] = update_ret["result"]
            else:
                for key, value in update_payload["ret"].items():
                    plan_state[key] = value

        if not result["result"]:
            result["comment"] = result["comment"] + (f"{name} already exists",)
    else:
        delivery_props = {}
        dc_payload = {"name": name, "s3BucketName": s3_bucket_name}
        if sns_topic_arn:
            dc_payload["snsTopicARN"] = sns_topic_arn
        if s3_key_prefix:
            dc_payload["s3KeyPrefix"] = s3_key_prefix
        if s3_kms_key_arn:
            dc_payload["s3KmsKeyArn"] = s3_kms_key_arn
        if config_snapshot_delivery_properties:
            delivery_props["deliveryFrequency"] = config_snapshot_delivery_properties[
                "delivery_frequency"
            ]
            dc_payload["configSnapshotDeliveryProperties"] = delivery_props
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=dc_payload
            )
            result["comment"] = (f"Would create aws.config.delivery_channel '{name}'",)
            return result

        ret = await hub.exec.boto3.client.config.put_delivery_channel(
            ctx, DeliveryChannel=dc_payload
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"] + ret["comment"]
            return result
            # resource_id = name
        result["comment"] = result["comment"] + (
            f"Created aws.config.delivery_channel '{name}'",
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
        if not resource_updated:
            result["comment"] = f"Would create aws.config.delivery_channel '{name}'"
        else:
            result["comment"] = (f"Would update aws.config.delivery_channel '{name}'",)
    elif (not before) or resource_updated:
        after = await hub.exec.boto3.client.config.describe_delivery_channels(
            ctx, DeliveryChannelNames=[resource_id]
        )
        result[
            "new_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=after["ret"]["DeliveryChannels"][0],
            idem_resource_name=name,
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Deletes the specified delivery channel Before you can delete the delivery channel, you must stop the configuration recorder
    by using the StopConfigurationRecorder action.

    Args:
        name(Text): An Idem name of the rule.
        resource_id(Text, optional): AWS Config delivery channel Name. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.delivery_channel.absent:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_id = resource_id if resource_id else name
    # if not resource_id:
    #     result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
    #         resource_type="aws.config.delivery_channel", name=name
    #     )
    #     return result
    before = await hub.exec.boto3.client.config.describe_delivery_channels(
        ctx, DeliveryChannelNames=[resource_id]
    )
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.delivery_channel", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx,
            raw_resource=before["ret"]["DeliveryChannels"][0],
            idem_resource_name=resource_id,
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.delivery_channel", name=resource_id
        )
        return result
    else:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
                ctx,
                raw_resource=before["ret"]["DeliveryChannels"][0],
                idem_resource_name=resource_id,
            )

            ret = await hub.exec.boto3.client.config.delete_delivery_channel(
                ctx, DeliveryChannelName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.config.delivery_channel", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Return details about your delivery channel.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.config.delivery_channel
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_delivery_channels(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe Delivery Channels {ret['comment']}")
        return {}

    for resource in ret["ret"]["DeliveryChannels"]:
        resource_id = resource.get("name")
        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_delivery_channel_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.config.delivery_channel.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.efs.create_mount_target
hub.exec.boto3.client.efs.delete_mount_target
hub.exec.boto3.client.efs.describe_mount_targets
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    file_system_id: str,
    subnet_id: str,
    resource_id: str = None,
    ip_address: str = None,
    security_groups: List[str] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a mount target for a file system. You can then mount the file system on EC2 instances by using the mount
    target. You can create one mount target in each Availability Zone in your VPC. All EC2 instances in a VPC within
    a given Availability Zone share a single mount target for a given file system. If you have multiple subnets in
    an Availability Zone, you create a mount target in one of the subnets. EC2 instances do not need to be in the
    same subnet as the mount target in order to access their file system. You can create only one mount target for
    an EFS file system using One Zone storage classes. You must create that mount target in the same Availability
    Zone in which the file system is located. Use the AvailabilityZoneName and AvailabiltyZoneId properties in the
    DescribeFileSystems response object to get this information. Use the subnetId associated with the file system's
    Availability Zone when creating the mount target. For more information, see Amazon EFS: How it Works.  To create
    a mount target for a file system, the file system's lifecycle state must be available. For more information, see
    DescribeFileSystems. In the request, provide the following:   The file system ID for which you are creating the
    mount target.   A subnet ID, which determines the following:   The VPC in which Amazon EFS creates the mount
    target   The Availability Zone in which Amazon EFS creates the mount target   The IP address range from which
    Amazon EFS selects the IP address of the mount target (if you don't specify an IP address in the request)
    After creating the mount target, Amazon EFS returns a response that includes, a MountTargetId and an IpAddress.
    You use this IP address when mounting the file system in an EC2 instance. You can also use the mount target's
    DNS name when mounting the file system. The EC2 instance on which you mount the file system by using the mount
    target can resolve the mount target's DNS name to its IP address. For more information, see How it Works:
    Implementation Overview.  Note that you can create mount targets for a file system in only one VPC, and there
    can be only one mount target per Availability Zone. That is, if the file system already has one or more mount
    targets created for it, the subnet specified in the request to add another mount target must meet the following
    requirements:   Must belong to the same VPC as the subnets of the existing mount targets   Must not be in the
    same Availability Zone as any of the subnets of the existing mount targets   If the request satisfies the
    requirements, Amazon EFS does the following:   Creates a new mount target in the specified subnet.   Also
    creates a new network interface in the subnet as follows:   If the request provides an IpAddress, Amazon EFS
    assigns that IP address to the network interface. Otherwise, Amazon EFS assigns a free address in the subnet (in
    the same way that the Amazon EC2 CreateNetworkInterface call does when a request does not specify a primary
    private IP address).   If the request provides SecurityGroups, this network interface is associated with those
    security groups. Otherwise, it belongs to the default security group for the subnet's VPC.   Assigns the
    description Mount target fsmt-id for file system fs-id  where  fsmt-id  is the mount target ID, and  fs-id  is
    the FileSystemId.   Sets the requesterManaged property of the network interface to true, and the requesterId
    value to EFS.   Each Amazon EFS mount target has one corresponding requester-managed EC2 network interface.
    After the network interface is created, Amazon EFS sets the NetworkInterfaceId field in the mount target's
    description to the network interface ID, and the IpAddress field to its address. If network interface creation
    fails, the entire CreateMountTarget operation fails.    The CreateMountTarget call returns only after creating
    the network interface, but while the mount target state is still creating, you can check the mount target
    creation status by calling the DescribeMountTargets operation, which among other things returns the mount target
    state.  We recommend that you create a mount target in each of the Availability Zones. There are cost
    considerations for using a file system in an Availability Zone through a mount target created in another
    Availability Zone. For more information, see Amazon EFS. In addition, by always using a mount target local to
    the instance's Availability Zone, you eliminate a partial failure scenario. If the Availability Zone in which
    your mount target is created goes down, then you can't access your file system through that mount target.  This
    operation requires permissions for the following action on the file system:
    elasticfilesystem:CreateMountTarget    This operation also requires permissions for the following Amazon EC2
    actions:    ec2:DescribeSubnets     ec2:DescribeNetworkInterfaces     ec2:CreateNetworkInterface

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): The ID of mount target. Defaults to None.
        file_system_id(Text): The ID of the file system for which to create the mount target.
        subnet_id(Text): The ID of the subnet to add the mount target in. For file systems that use One Zone storage
            classes, use the subnet that is associated with the file system's Availability Zone.
        ip_address(Text, optional): Valid IPv4 address within the address range of the specified subnet. Defaults to None.
        security_groups(List[str], optional): Up to five VPC security group IDs, of the form sg-xxxxxxxx. These must be for the same VPC as
            subnet specified. Defaults to None.

    Request Syntax:
       [mount-target-resource-id]:
            aws.efs.mount_target.present:
            - name: 'string'
            - resource_id: 'string'
            - security_groups: List
            - file_system_id: 'string'
            - subnet_id: 'string'
            - ip_address: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            fsmt-xxxxxxxxxxx:
              aws.efs.mount_target.present:
              - name: fsmt-xxxxxxxxxxx
              - resource_id: fsmt-xxxxxxxxxxx
              - security_groups:
                - sg-xxxxxxxxx
                - sg-xxxxxxxxx
              - file_system_id: fs-xxxxxxxxxxxxxxx
              - subnet_id: subnet-xxxxxxxxxx
              - ip_address: 127.0.0.1
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for mount target
    mount_target = None
    if resource_id:
        mount_target = await hub.exec.boto3.client.efs.describe_mount_targets(
            ctx, MountTargetId=resource_id
        )
        if not mount_target["result"]:
            result["comment"] = mount_target["comment"]
            result["result"] = mount_target["result"]
            return result

    # Check for existing mount target
    before = None
    if mount_target and mount_target["ret"]["MountTargets"]:
        before = mount_target["ret"]["MountTargets"][0]

    # Check for security groups of mount target
    mount_target_security_groups = []
    if before:
        security_groups_result = (
            await hub.exec.boto3.client.efs.describe_mount_target_security_groups(
                ctx, MountTargetId=resource_id
            )
        )
        if not security_groups_result["result"]:
            hub.log.debug(
                f"Could not describe security groups for mount target {resource_id} : {security_groups_result['comment']}"
            )
            result["comment"] = (security_groups_result["comment"],)
            result["result"] = False
            return result

        mount_target_security_groups = security_groups_result["ret"]["SecurityGroups"]
    # Update current state
    current_state = (
        hub.tool.aws.efs.mount_target_utils.convert_raw_mount_target_to_present(
            ctx=ctx,
            mount_target=before,
            security_groups=mount_target_security_groups,
        )
    )
    result["old_state"] = current_state

    # Handle No change behaviour
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "file_system_id": file_system_id,
        "subnet_id": subnet_id,
        "ip_address": ip_address,
        "security_groups": security_groups,
    }

    is_security_groups_updated = not current_state or (
        not hub.tool.aws.state_comparison_utils.are_lists_identical(
            current_state.get("security_groups"), desired_state.get("security_groups")
        )
    )

    is_change_detected = before is None or is_security_groups_updated

    if not is_change_detected:
        result["comment"] = (f"aws.efs.mount_target '{name}' already exists",)
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.efs.mount_target", name=name
            )
            if before
            else hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.efs.mount_target", name=name
            )
        )
        return result

    # Handle actual resource create or update
    if before:
        ret = await hub.exec.boto3.client.efs.modify_mount_target_security_groups(
            ctx, MountTargetId=resource_id, SecurityGroups=security_groups
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = (
            f"Updated security groups for aws.efs.mount_target '{name}'",
        )
    else:
        ret = await hub.exec.boto3.client.efs.create_mount_target(
            ctx,
            **{
                "FileSystemId": file_system_id,
                "SubnetId": subnet_id,
                "IpAddress": ip_address,
                "SecurityGroups": security_groups,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.efs.mount_target", name=name
        )
        resource_id = ret["ret"]["MountTargetId"]

    # Fetch the updated resource and update new_state
    mount_target = await hub.exec.boto3.client.efs.describe_mount_targets(
        ctx, MountTargetId=resource_id
    )
    if not mount_target["result"]:
        result["comment"] = result["comment"] + mount_target["comment"]
        result["result"] = mount_target["result"]
        return result

    after = mount_target["ret"]["MountTargets"][0]
    security_groups_result = (
        await hub.exec.boto3.client.efs.describe_mount_target_security_groups(
            ctx, MountTargetId=resource_id
        )
    )
    mount_target_security_groups = security_groups_result["ret"]["SecurityGroups"]
    result[
        "new_state"
    ] = hub.tool.aws.efs.mount_target_utils.convert_raw_mount_target_to_present(
        ctx=ctx,
        mount_target=after,
        security_groups=mount_target_security_groups,
    )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified mount target. This operation forcibly breaks any mounts of the file system by using the
    mount target that is being deleted, which might disrupt instances or applications using those mounts. To avoid
    applications getting cut off abruptly, you might consider unmounting any mounts of the mount target, if
    feasible. The operation also deletes the associated network interface. Uncommitted writes might be lost, but
    breaking a mount target using this operation does not corrupt the file system itself. The file system you
    created remains. You can mount an EC2 instance in your VPC by using another mount target. This operation
    requires permissions for the following action on the file system:    elasticfilesystem:DeleteMountTarget     The
    DeleteMountTarget call returns while the mount target state is still deleting. You can check the mount target
    deletion by calling the DescribeMountTargets operation, which returns a list of mount target descriptions for
    the given file system.   The operation also requires permissions for the following Amazon EC2 action on the
    mount target's network interface:    ec2:DeleteNetworkInterface

    Args:
        name(Text, optional): An Idem name of the resource.
        resource_id(Text, optional): The ID of mount target.
        timeout(Dict, optional): Timeout configuration for deletion of AWS EFS mount target.
            * delete (Dict) -- Timeout configuration for deletion of an EFS mount target
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40

    Request Syntax:
       [mount-target-resource-id]:
            aws.efs.mount_target.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            fsmt-xxxxxxxxxxx:
              aws.efs.mount_target.absent:
              - name: fsmt-xxxxxxxxxxx
              - resource_id: fsmt-xxxxxxxxxxx
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for mount target
    mount_target = None
    if resource_id:
        mount_target = await hub.exec.boto3.client.efs.describe_mount_targets(
            ctx, MountTargetId=resource_id
        )

    # Check for existing mount target
    before = None
    if mount_target and mount_target["result"] and mount_target["ret"]["MountTargets"]:
        before = mount_target["ret"]["MountTargets"][0]

    if before:
        # Check for mount target security group
        mount_target_security_groups = []
        security_groups_result = (
            await hub.exec.boto3.client.efs.describe_mount_target_security_groups(
                ctx, MountTargetId=resource_id
            )
        )
        if security_groups_result["result"]:
            mount_target_security_groups = security_groups_result["ret"][
                "SecurityGroups"
            ]
        else:
            hub.log.debug(
                f"Could not describe security groups for mount target {resource_id} : {security_groups_result['comment']}"
            )
        result[
            "old_state"
        ] = hub.tool.aws.efs.mount_target_utils.convert_raw_mount_target_to_present(
            ctx=ctx,
            mount_target=before,
            security_groups=mount_target_security_groups,
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.efs.mount_target", name=name
        )
    elif ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.efs.mount_target", name=name
        )
    else:
        ret = await hub.exec.boto3.client.efs.delete_mount_target(
            ctx, MountTargetId=resource_id
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        # Custom waiter for delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        delete_waiter_acceptors = [
            {
                "matcher": "error",
                "expected": "MountTargetNotFound",
                "state": "success",
                "argument": "Error.Code",
            },
            {
                "matcher": "pathAll",
                "expected": "error",
                "state": "success",
                "argument": "MountTargets[].LifeCycleState",
            },
            {
                "matcher": "pathAll",
                "expected": "available",
                "state": "retry",
                "argument": "MountTargets[].LifeCycleState",
            },
            {
                "matcher": "pathAll",
                "expected": "deleting",
                "state": "retry",
                "argument": "MountTargets[].LifeCycleState",
            },
            {
                "matcher": "pathAll",
                "expected": "deleted",
                "state": "success",
                "argument": "MountTargets[].LifeCycleState",
            },
        ]

        cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="MountTargetDelete",
            operation="DescribeMountTargets",
            argument=["Error.Code", "MountTargets[].LifeCycleState"],
            acceptors=delete_waiter_acceptors,
            client=hub.tool.boto3.client.get_client(ctx, "efs"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "efs",
                "MountTargetDelete",
                cluster_waiter,
                MountTargetId=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.efs.mount_target", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns the descriptions of all the current mount targets, or a specific mount target, for a file system. When
    requesting all of the current mount targets, the order of mount targets returned in the response is unspecified.
    This operation requires permissions for the elasticfilesystem:DescribeMountTargets action, on either the file
    system ID that you specify in FileSystemId, or on the file system of the mount target that you specify in
    MountTargetId.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.efs.mount_target
    """
    result = {}

    # Fetch all file system ids
    file_system_ids = []
    file_systems = await hub.exec.boto3.client.efs.describe_file_systems(ctx)
    if not file_systems["result"]:
        hub.log.debug(f"Could not describe file_system {file_systems['comment']}")
        return result
    for file_system in file_systems["ret"]["FileSystems"]:
        file_system_ids.append(file_system["FileSystemId"])

    # Fetch all mount targets
    mount_targets = []
    for file_system_id in file_system_ids:
        mount_targets_result = await hub.exec.boto3.client.efs.describe_mount_targets(
            ctx, FileSystemId=file_system_id
        )
        if not mount_targets_result["result"]:
            hub.log.debug(
                f"Could not describe mount_target for file system {file_system_id} : {mount_targets_result['comment']}"
            )
            continue
        mount_targets.extend(mount_targets_result["ret"]["MountTargets"])

    # Describe each mount target
    for mount_target in mount_targets:
        mount_target_id = mount_target["MountTargetId"]

        security_groups_result = (
            await hub.exec.boto3.client.efs.describe_mount_target_security_groups(
                ctx, MountTargetId=mount_target_id
            )
        )
        if not security_groups_result["result"]:
            hub.log.debug(
                f"Could not describe security groups for mount target {mount_target_id} : {security_groups_result['comment']}"
            )
            continue

        security_groups = security_groups_result["ret"]["SecurityGroups"]
        resource = (
            hub.tool.aws.efs.mount_target_utils.convert_raw_mount_target_to_present(
                ctx=ctx,
                mount_target=mount_target,
                security_groups=security_groups,
            )
        )
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "aws.efs.mount_target.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""

Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_security_group
hub.exec.boto3.client.ec2.delete_security_group
hub.exec.boto3.client.ec2.describe_security_groups
hub.tool.boto3.resource.exec(resource, authorize_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, authorize_ingress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_egress, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, revoke_ingress, *args, **kwargs)
"""
import copy
from copy import deepcopy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.ec2.vpc.present"]},
    "absent": {
        "require": ["aws.eks.cluster.absent", "aws.ec2.security_group_rule.absent"]
    },
}


async def present(
    hub,
    ctx,
    name: str,
    description: str,
    vpc_id: str,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a security group. A security group acts as a virtual firewall for your instance to control inbound and
    outbound traffic. For more information, see Amazon EC2 security groups in the Amazon Elastic Compute Cloud User
    Guide and Security groups for your VPC in the Amazon Virtual Private Cloud User Guide. When you create a
    security group, you specify a friendly name of your choice. You can have a security group for use in EC2-Classic
    with the same name as a security group for use in a VPC. However, you can't have two security groups for use in
    EC2-Classic with the same name or two security groups for use in a VPC with the same name. You have a default
    security group for use in EC2-Classic and a default security group for use in your VPC. If you don't specify a
    security group when you launch an instance, the instance is launched into the appropriate default security
    group. A default security group includes a default rule that grants instances unrestricted network access to
    each other. You can add or remove rules from your security groups using AuthorizeSecurityGroupIngress,
    AuthorizeSecurityGroupEgress, RevokeSecurityGroupIngress, and RevokeSecurityGroupEgress. For more information
    about VPC security group limits, see Amazon VPC Limits.

    Args:
        name(Text): The security group name as set on AWS.
        description(Text): Description of the security group.
        vpc_id(Text): Id of the VPC security group should be attached to.
        resource_id(Text, optional): AWS Security Group ID.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the security group.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [security_group_id]:
          aws.ec2.security_group.present:
          - resource_id: 'string'
          - name: 'string'
          - vpc_id: 'string'
          - tags:
                - Key: 'string'
                  Value: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d:
              aws.ec2.security_group.present:
                  - resource_id: sg-0008bd25b7867b5cf
                  - name: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                  - vpc_id: vpc-247e9a5d
                  - description: sg-description
                  - tags:
                        - Key: Name
                          Value: sg-name
                        - Key: sg-tag-key-2
                          Value: sg-tag-value-2
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "ec2", "SecurityGroup", resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)
    if before:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
            plan_state = copy.deepcopy(result["old_state"])
            old_tags = result["old_state"].get("tags")
            resource_updated = False
            if tags is not None and tags != old_tags:
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx,
                    resource_id=resource_id,
                    old_tags=old_tags,
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + (update_ret["comment"])
                    result["result"] = update_ret["result"]
                    return result
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or update_ret["result"]
                if ctx.get("test", False) and resource_updated:
                    plan_state["tags"] = update_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update aws.ec2.security_group {name}",
                    )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "description": description,
                    "vpc_id": vpc_id,
                    "tags": tags,
                },
            )
            result["comment"] = result["comment"] + (
                f"Would create aws.ec2.security_group {name}",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.create_security_group(
                ctx,
                GroupName=name,
                Description=description,
                VpcId=vpc_id,
                TagSpecifications=[
                    {
                        "ResourceType": "security-group",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags is not None
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + (ret["comment"])
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_id = ret["ret"]["GroupId"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(
                ctx, "ec2", "SecurityGroup", resource_id
            )
            after = await hub.tool.boto3.resource.describe(resource)

            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(after)
        else:
            result["new_state"] = deepcopy(result["old_state"])

    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a security group. If you attempt to delete a security group that is associated with an instance, or is
    referenced by another security group, the operation fails with InvalidGroup.InUse in EC2-Classic or
    DependencyViolation in EC2-VPC.

    Args:
        name(Text): An Idem name to identify the security group resource.
        resource_id(Text, optional): AWS Security Group ID. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            sg-0008bd25b7867b5cf:
              aws.ec2.security_group.absent:
                - name: isolation-securitygroup-84c9a893-844b-40ab-86e8-b565dea88b5d
                - resource_id: id: sg-0008bd25b7867b5cf
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.security_group", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "SecurityGroup", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = result[
            "comment"
        ] + hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.security_group", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.security_group", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(before)
        try:
            ret = await hub.exec.boto3.client.ec2.delete_security_group(
                ctx, GroupId=resource_id
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + (ret["comment"])
                result["result"] = ret["result"]
                return result
            result["comment"] = result["comment"] + (
                f"Deleted aws.ec2.security_group '{name}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes the specified security groups or all of your security groups. A security group is for use with
    instances either in the EC2-Classic platform or in a specific VPC. For more information, see Amazon EC2 security
    groups in the Amazon Elastic Compute Cloud User Guide and Security groups for your VPC in the Amazon Virtual
    Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.security_group
    """
    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_security_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe security_group {ret['comment']}")
        return {}

    for security_group in ret["ret"]["SecurityGroups"]:
        translated_security_group = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(security_group)
        )

        result[security_group.get("GroupId")] = {
            "aws.ec2.security_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_security_group.items()
            ]
        }
    return result


async def search(
    hub,
    ctx,
    name,
    filters: List = None,
    resource_id: str = None,
    vpc_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    r"""
    Provides details about a specific Security Group as a data-source. Supply one of the inputs as the filter.

    Args:
        name(Text):
             The name of the Idem state.

        resource_id(Text, optional):
            Security Group id to identify the resource.

        filters(list, optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_security_groups

        vpc_id(Text, Optional): Id of the VPC security group attached to.

        tags(List or Dict, Optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value} for security group.

    Request Syntax:
        [Idem-state-name]:
          aws.ec2.security_group.search:
              - name: 'string'
              - resource_id: 'string'
              - filters:
                - name: 'string'
                  values: 'list'
                - name: 'string'
                  values: 'list'
              - vpc_id: 'string'
              - tags:
                 - Key: 'string'
                   Value: 'string'
    Examples:

        .. code-block:: bash

            idem-test-security-group-search:
                aws.ec2.security_group.search:
                  - name: idem-test-security-group
                  - filters:
                    - name: 'group-name'
                      values: ["idem-test-security-group"]
                  - vpc_id: vpc-247e9a5d
                  - tags:
                    - Key: Name
                      Value: sg-name
                    - Key: sg-tag-key-2
                      Value: sg-tag-value-2


    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["comment"] = syntax_validation["comment"]
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    if isinstance(tags, Dict):
        tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
    if tags:
        for tag in tags:
            boto3_filter.append({"Name": "tag:" + tag["Key"], "Values": [tag["Value"]]})
    if vpc_id:
        boto3_filter.append({"Name": "vpc-id", "Values": [vpc_id]})

    security_groups_args = {}
    if boto3_filter:
        security_groups_args["Filters"] = boto3_filter
    # Including None as GroupIds, throws ParamValidationError for describe_security_groups
    if resource_id:
        security_groups_args["GroupIds"] = [resource_id]

    ret = await hub.exec.boto3.client.ec2.describe_security_groups(
        ctx, **security_groups_args
    )

    if not ret["result"]:
        result["result"] = False
        result["comment"] = ret["comment"]
        return result

    if not ret["ret"]["SecurityGroups"]:
        result["comment"] = (f"Unable to find aws.ec2.security_group '{name}'",)
        return result

    security_group = ret["ret"]["SecurityGroups"][0]
    if len(ret["ret"]["SecurityGroups"]) > 1:
        result["comment"] = (
            f"More than one aws.ec2.security_group resource was found. Use resource {security_group.get('GroupId')}",
        )
    result["old_state"] = hub.tool.aws.ec2.conversion_utils.convert_raw_sg_to_present(
        security_group
    )

    # Populate both "old_state" and "new_state" with the same data
    result["new_state"] = copy.deepcopy(result["old_state"])
    return result



import json
from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_instance_profile_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "InstanceProfileName": "name",
            "InstanceProfileId": "id",
            "Path": "path",
        }
    )

    # Name is the unique identifier for instance_profile so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("InstanceProfileName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    translated_resource["roles"] = []
    for role in raw_resource.get("Roles"):
        translated_resource["roles"].append({"RoleName": role.get("RoleName")})

    return translated_resource


async def convert_raw_policy_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "PolicyName": "name",
            "Arn": "resource_id",
            "PolicyId": "id",
            "Path": "path",
            "Description": "description",
            "Tags": "tags",
        }
    )

    policy_arn = raw_resource.get("Arn")
    # Arn is the unique identifier for policy so it is set as resource_id
    translated_resource = {}

    default_version_id = raw_resource.get("DefaultVersionId")
    ret_get = await hub.exec.boto3.client.iam.get_policy(ctx, PolicyArn=policy_arn)
    if not ret_get["result"]:
        hub.log.warning(
            f"Failed on fetching policy with arn {policy_arn} with error {ret_get['comment']}."
        )
        return None

    ret_get_version = await hub.exec.boto3.client.iam.get_policy_version(
        ctx, PolicyArn=policy_arn, VersionId=default_version_id
    )
    if not ret_get_version["result"]:
        hub.log.warning(
            f"Failed on fetching policy document with arn {policy_arn} and version {default_version_id} "
            f"with error {ret_get_version['comment']} Describe will skip this policy and continue."
        )
        return None

    document = _standardise_json(
        ret_get_version["ret"]["PolicyVersion"].get("Document")
    )
    translated_resource["policy_document"] = document
    translated_resource["default_version_id"] = default_version_id

    # Get policy tags
    if raw_resource.get("Tags", None) is None:
        ret_get_tags = await hub.exec.boto3.client.iam.list_policy_tags(
            ctx, PolicyArn=policy_arn
        )
        if ret_get_tags["result"]:
            translated_resource["tags"] = ret_get_tags["ret"].get("Tags", None)

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if translated_resource.get("tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            translated_resource.get("tags")
        )
    return translated_resource


def convert_raw_role_to_present(hub, raw_resource: Dict[str, Any]) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "RoleName": "name",
            "Arn": "arn",
            "RoleId": "id",
            "Path": "path",
            "Description": "description",
            "MaxSessionDuration": "max_session_duration",
        }
    )

    # RoleName is the unique identifier for policy so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("RoleName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    if raw_resource.get("PermissionsBoundary"):
        translated_resource["permissions_boundary"] = raw_resource.get(
            "PermissionsBoundary"
        ).get("PermissionsBoundaryArn")

    if raw_resource.get("AssumeRolePolicyDocument", None):
        translated_resource["assume_role_policy_document"] = _standardise_json(
            raw_resource.get("AssumeRolePolicyDocument")
        )

    return translated_resource


def convert_raw_role_policy_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    # Policy name is the 'name' and 'resource_id' of the role_policy
    resource_parameters = OrderedDict(
        {
            "RoleName": "role_name",
            "PolicyName": "name",
        }
    )

    # Use {role_name}-{policy_name} as resource_id
    translated_resource = {
        "resource_id": f"{raw_resource.get('RoleName')}-{raw_resource.get('PolicyName')}"
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("PolicyDocument"):
        translated_resource["policy_document"] = _standardise_json(
            raw_resource.get("PolicyDocument")
        )

    return translated_resource


def convert_raw_role_policy_attachment_to_present(
    hub, role_name: str, policy_arn: str
) -> Dict[str, Any]:

    translated_resource = {"role_name": role_name, "policy_arn": policy_arn}

    return translated_resource


def convert_raw_user_policy_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    # Policy name is the 'name' and 'resource_id' of the user_policy
    resource_parameters = OrderedDict(
        {
            "UserName": "user_name",
            "PolicyName": "name",
        }
    )

    # Use {user_name}-{policy_name} as resource_id
    translated_resource = {
        "resource_id": f"{raw_resource.get('UserName')}-{raw_resource.get('PolicyName')}"
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("PolicyDocument"):
        translated_resource["policy_document"] = _standardise_json(
            raw_resource.get("PolicyDocument")
        )

    return translated_resource


def convert_raw_user_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    # Arn is not used for present but required for arg binding
    resource_parameters = OrderedDict(
        {
            "Arn": "arn",
            "Path": "path",
            "PermissionsBoundary": "permissions_boundary",
            "UserName": "user_name",
        }
    )
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": raw_resource.get("UserName"),
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    return translated_resource


def convert_raw_oidc_provider_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_id: str,
    idem_resource_name: str,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "Url": "url",
            "ClientIDList": "client_id_list",
            "ThumbprintList": "thumbprint_list",
        }
    )
    translated_resource = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)
    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    translated_resource["url"] = "https://" + raw_resource.get("Url")
    return translated_resource


def convert_raw_user_ssh_key_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str,
    ret_ssh_public_key_body: Dict[str, Any],
) -> Dict[str, Any]:
    translated_resource = {
        "name": idem_resource_name,
        "resource_id": idem_resource_name,
    }
    describe_parameters = OrderedDict(
        {
            "UserName": "user_name",
            "Status": "status",
        }
    )
    for (
        parameter_old_key,
        parameter_new_key,
    ) in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            translated_resource[parameter_new_key] = raw_resource.get(parameter_old_key)
    if ret_ssh_public_key_body["ret"].get("SSHPublicKey"):
        translated_resource["ssh_public_key_body"] = ret_ssh_public_key_body["ret"][
            "SSHPublicKey"
        ].get("SSHPublicKeyBody")
    else:
        hub.log.warning(
            f"Could not get ssh public key body with ssh public key ID {idem_resource_name} with error"
        )
    return translated_resource


def _standardise_json(value: str or Dict):
    if value is None:
        return None

    if isinstance(value, str):
        json_dict = json.loads(value)
    else:
        json_dict = value
    return json.dumps(json_dict, separators=(", ", ": "), sort_keys=True)


# NOTE: access_keys are different from other AWS objects, so we
#  use convert_raw_access_key_to_camel_case to provide a semi-normal python representation
#  and then convert_access_key_to_present takes that camel case representation
#  and outputs in 'present' format.
def convert_raw_access_key_to_camel_case(
    hub,
    access_key_raw: Dict[str, Any],
) -> Dict[str, Any]:
    # https://docs.aws.amazon.com/IAM/latest/APIReference/API_AccessKey.html
    # https://docs.aws.amazon.com/IAM/latest/APIReference/API_AccessKeyMetadata.html
    # All fields are optional, including the "access_key_id"
    parameter_map = OrderedDict(
        {
            "AccessKeyId": "access_key_id",
            "CreateDate": "create_date",
            "SecretAccessKey": "secret_access_key",  # only returned on creation
            "Status": "status",
            "UserName": "user_name",
        }
    )

    access_key = {}
    for api_k, new_k in parameter_map.items():
        if api_k in access_key_raw:
            access_key[new_k] = access_key_raw[api_k]

    return access_key


def convert_access_key_to_present(hub, access_key, idem_resource_name: str):
    # this takes as input the output of convert_raw_access_key_to_camel_case
    parameter_map = OrderedDict(
        {
            "access_key_id": "resource_id",
            "secret_access_key": "secret_access_key",
            "status": "status",
            "user_name": "user_name",
        }
    )
    translated_resource = {"name": idem_resource_name}
    for api_k, new_k in parameter_map.items():
        if api_k in access_key:
            translated_resource[new_k] = access_key[api_k]

    return translated_resource



from typing import Dict
from typing import List


def get_hosted_zone_with_filters(
    hub,
    raw_hosted_zones: Dict,
    hosted_zone_name: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: List = None,
):
    """
    Returns the hosted_zone with the specified filters, if it is available.

    Args:
        raw_hosted_zones(Dict): Dict of all the described hosted_zones
        hosted_zone_name(string, optional): Domain name of hosted_zone to filter.
        private_zone(bool, optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(string, optional): The vpc_id associated with the hosted_zone. One of the filter option for hosted_zone
        tags(List, optional): Tags of the hosted_zone. One of the filter option for hosted_zone

    """
    result = dict(comment=(), result=True, ret=None)
    hosted_zones = raw_hosted_zones["HostedZones"]

    # filter_hosted_zones() returns True if all the filters match for a hosted_zone, and it is added to the list
    filtered_hosted_zones = list(
        filter(
            lambda x: filter_hosted_zones(
                x, hosted_zone_name, private_zone, vpc_id, tags
            ),
            hosted_zones,
        )
    )

    if not filtered_hosted_zones:
        result["comment"] = (
            f"Unable to find aws.route53.hosted_zone resource with given filters",
        )
        return result
    resource_id = filtered_hosted_zones[0].get("Id").split("/")[-1]
    if len(filtered_hosted_zones) > 1:
        result["comment"] = (
            f"More than one aws.route53.hosted_zone resource was found with given filters. Use resource {resource_id}",
        )
    else:
        result["comment"] = (
            f"Found this aws.route53.hosted_zone resource {resource_id} with given filters",
        )
    # Building the format in which hosted_zone conversion_utils takes the hosted_zone resource
    res_hosted_zone = {
        "ret": {
            "HostedZone": filtered_hosted_zones[0],
            "VPCs": filtered_hosted_zones[0].get("VPCs"),
            "DelegationSet": filtered_hosted_zones[0].get("DelegationSet"),
        },
        "tags": filtered_hosted_zones[0].get("Tags"),
    }
    result["ret"] = res_hosted_zone
    return result


def filter_hosted_zones(
    hosted_zone: Dict,
    hosted_zone_name: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: List = None,
):
    """
    Returns True if the hosted_zone checks all the filters provided or return False

    Args:
        hosted_zone(Dict): The described hosted_zone
        hosted_zone_name(string, optional): Domain name of hosted_zone to filter.
        private_zone(bool, optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(string, optional): The vpc_id associated with the hosted_zone. One of the filter option for hosted_zone
        tags(List, optional): Tags of the hosted_zone. One of the filter option for hosted_zone

    """
    # Return True if all the provided filters match or return False.

    if hosted_zone_name:
        if hosted_zone["Name"] != hosted_zone_name:
            return False

    if private_zone is not None:
        if hosted_zone["Config"]["PrivateZone"] != private_zone:
            return False

    if vpc_id:
        found = False
        if hosted_zone["VPCs"]:
            for vpc in hosted_zone["VPCs"]:
                if vpc["VPCId"] == vpc_id:
                    found = True
                    break
            if not found:
                return False

    # Checking if all the tags in the filter match with the tags present in the hosted_zone.If not we return False
    if tags:
        tags2 = hosted_zone.get("Tags")
        if tags2 is None:
            return False
        tags2_map = {tag.get("Key"): tag for tag in tags2}
        for tag in tags:
            if tag["Key"] not in tags2_map or (
                tags2_map.get(tag["Key"]).get("Value") != tag["Value"]
            ):
                return False

    return True



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_service_linked_role_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "RoleName": "name",
            "Description": "description",
        }
    )

    # RoleName is the unique identifier for policy so it is set as resource_id
    translated_resource = {"resource_id": raw_resource.get("RoleName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            translated_resource[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("Tags") is not None:
        translated_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource.get("Tags")
        )

    name = raw_resource.get("RoleName")
    # If custom suffix was added in service-linked role name,
    # Then format of role name will be like -> DefaultServiceLinkedRoleName + "_" + custom_suffix
    if "_" in name:
        translated_resource["custom_suffix"] = name[name.index("_") + 1 :]
    translated_resource["service_name"] = raw_resource.get("Arn").split("/")[-2]

    return translated_resource



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

AWS_Org_Not_In_Use_Exception = "AWSOrganizationsNotInUseException"
ALREADY_EXISTS = "already exists"

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.organization_unit.absent",
        ],
    },
}


async def present(hub, ctx, name: str, feature_set: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an AWS organization. The account whose user is calling the CreateOrganization operation
    automatically becomes the management account of the new organization.This operation must be called using
    credentials from the account that is to become the new organization's management account. The principal
    must also have the relevant IAM permissions.By default (or if you set the FeatureSet parameter to ALL ),
    the new organization is created with all features enabled and service control policies automatically
    enabled in the root. If you instead choose to create the organization supporting only the
    consolidated billing features by setting the FeatureSet parameter to CONSOLIDATED_BILLING" ,
    no policy types are enabled by default, and you can't use organization policies

    Args:
        name(Text): Name of the resource.
        feature_set(Text, Default: 'ALL'): optional values: 'ALL'|'CONSOLIDATED_BILLING'

    Request Syntax:
        [org-id]:
          aws.organizations.organization.present:
          - feature_set: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            org_is_present:
              aws.organizations.organization.present:
                - feature_set: 'ALL'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organization(ctx)

    if not before:
        # Org not present , create

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "feature_set": feature_set,
                },
            )
            result["comment"] = result["comment"] + (
                f"Would create aws.organizations.organization '{name}'.",
            )
            return result

        try:
            ret = await hub.exec.boto3.client.organizations.create_organization(
                ctx, FeatureSet=feature_set
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result["comment"] + (
                f"Created aws.organizations.organization {name}.",
            )
            after = await hub.exec.boto3.client.organizations.describe_organization(ctx)
            if after and after.get("ret"):
                result[
                    "new_state"
                ] = hub.tool.aws.organization.conversion_utils.convert_raw_organization_to_present(
                    after["ret"]["Organization"], name
                )
                roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

                if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
                    result["new_state"]["roots"] = roots_resp["ret"]["Roots"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        # Org present
        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_organization_to_present(
            before["ret"]["Organization"], name
        )
        roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

        if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
            result["old_state"]["roots"] = roots_resp["ret"]["Roots"]
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization '{name}' {ALREADY_EXISTS}.",
        )

    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the organization. You can delete an organization only by using credentials from the
    management account. The organization must be empty of member accounts.

    Args:
        name(Text): Name of the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            aws.organizations.organization.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_organization(ctx)
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_organization_to_present(
            before["ret"], name
        )
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.organizations.organization '{name}'.",
            )
            return result

    if not before and AWS_Org_Not_In_Use_Exception in str(before["comment"]):
        result["comment"] = result["comment"] + (
            f"aws.organizations.organization '{name}' is already absent",
        )
        return result
    else:
        try:
            ret = await hub.exec.boto3.client.organizations.delete_organization(ctx)

            if not ret["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (
                f"aws.organizations.organization {name} deleted.",
            )
            result["result"] = ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    try:
        after = await hub.exec.boto3.client.organizations.describe_organization(ctx)
        if after.get("ret"):
            result[
                "new_state"
            ] = hub.tool.aws.organization.conversion_utils.convert_raw_organization_to_present(
                after["ret"], name
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Retrieves information about the organization that the user's account belongs to.
    This operation can be called from any account in the organization..


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.organization
    """

    result = {}
    ret = await hub.exec.boto3.client.organizations.describe_organization(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe organization with error:  {ret['comment']}")
        return {}

    resource = ret["ret"]["Organization"]

    translated_resource = (
        hub.tool.aws.organization.conversion_utils.convert_raw_organization_to_present(
            resource
        )
    )
    roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

    if roots_resp and roots_resp["ret"] and roots_resp["ret"]["Roots"]:
        translated_resource["roots"] = roots_resp["ret"]["Roots"]
    result[resource["Id"]] = {
        "aws.organizations.organization.present": [
            {parameter_key: parameter_value}
            for parameter_key, parameter_value in translated_resource.items()
        ]
    }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.delete_user_policy
hub.exec.boto3.client.iam.get_user_policy
hub.exec.boto3.client.iam.list_user_policies
hub.exec.boto3.client.iam.put_user_policy
resource = hub.tool.boto3.resource.create(ctx, "iam", "UserPolicy", name)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, put, *args, **kwargs)
"""
import asyncio
import copy
import re
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.iam.user_policy_attachment.absent",
        ],
    },
    "present": {
        "require": ["aws.iam.user.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    user_name: str,
    policy_document: Dict or str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Adds or updates an inline policy document that is embedded in the specified IAM user. An IAM user can also have
    a managed policy attached to it. To attach a managed policy to a user, use AttachUserPolicy. To create a new
    managed policy, use CreatePolicy. For information about policies, see Managed policies and inline policies in
    the IAM User Guide. For information about the maximum number of inline policies that you can embed in a user,
    see IAM and STS quotas in the IAM User Guide.  Because policy documents can be large, you should use POST rather
    than GET when calling PutUserPolicy. For general information about using the Query API with IAM, see Making
    query requests in the IAM User Guide.

    Args:
        name(Text): The name of the AWS IAM policy.
        user_name(Text): The UserPolicy's user_name identifier
        policy_document(dictionary): The policy document. IAM stores policies in JSON format. However, resources that were
         created using CloudFormation templates can be formatted in YAML. CloudFormation always converts a YAML policy to
         JSON format before submitting it to IAM.
        resource_id(Text, Optional): The user name and policy name with a separator '-'. Format: [user_name]-[policy_name]

    Request Syntax:
        [iam-user-policy-name]:
          aws.iam.user_policy.present:
          - resource_id: 'string'
          - name: 'string'
          - user_name: 'string'
          - policy_document: 'dict or string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-user-policy:
              aws.iam.user_policy.present:
                - resource_id: idem-test-user-idem-test-user-policy
                - name: idem-test-user-policy
                - user_name: idem-test-user
                - policy_document: '{"Version": "2012-10-17", "Statement": {"Effect": "Allow", "Action": ["cloudwatch:ListMetrics", "cloudwatch:GetMetricStatistics"], "Resource":"*"}}'

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    policy_name = name

    if resource_id:
        if not re.search(f"^({user_name})-({policy_name})$", resource_id):
            result[
                "comment"
            ] = f"Incorrect aws.iam.user_policy resource_id: {resource_id}. Expected id {user_name}-{policy_name}"
            result["result"] = False
            return result

    resource = hub.tool.boto3.resource.create(
        ctx,
        "iam",
        "UserPolicy",
        user_name=user_name,
        name=policy_name,
    )

    before = await hub.tool.boto3.resource.describe(resource)

    # Standardise on the json format, user policy should not be sorted
    policy_document = hub.tool.aws.state_comparison_utils.standardise_json(
        policy_document, sort_keys=False
    )

    if before:

        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_policy_to_present(before)

        if not hub.tool.aws.state_comparison_utils.is_json_identical(
            result["old_state"].get("policy_document"), policy_document
        ):

            if ctx.get("test", False):
                plan_state = copy.deepcopy(result["old_state"])
                plan_state["policy_document"] = policy_document
                result["new_state"] = plan_state
                result["comment"] = (
                    f"Would update aws.iam.user_policy '{policy_name}' for user '{user_name}'",
                )
                return result

            # call to update user-policy
            fret = await _update_user_policy(
                hub, ctx, user_name, policy_name, policy_document
            )
            result["result"] = fret["result"]
            if not result["result"]:
                result["comment"] = fret["comment"]
                return result

            result["comment"] = (
                f"Updated aws.iam.user_policy '{policy_name}' for user '{user_name}'",
            )

        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"] = (
                f"aws.iam.user_policy '{policy_name}' for user '{user_name}' already exists",
            )
            return result
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": policy_name,
                        "user_name": user_name,
                        "policy_document": policy_document,
                    },
                )
                result["comment"] = (
                    f"Would create aws.iam.user_policy '{policy_name}' for user '{user_name}'",
                )
                return result

            # call to create user-policy
            fret = await _update_user_policy(hub, ctx, user_name, name, policy_document)

            result["result"] = fret["result"]
            if not result["result"]:
                result["comment"] = fret["comment"]
                return result

            result["comment"] = (
                f"Created aws.iam.user_policy '{policy_name}' for user '{user_name}'",
            )

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result[
            "new_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_policy_to_present(after)
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, user_name: str = None, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified inline policy that is embedded in the specified IAM user. A user can also have managed
    policies attached to it. To detach a managed policy from a user, use DetachUserPolicy. For more information
    about policies, refer to Managed policies and inline policies in the IAM User Guide.

    Args:
        name(Text): The name of the AWS IAM policy.
        user_name(Text, optional): The UserPolicy's user_name identifier. Idem automatically considers this resource being absent
         if this field is not specified.
        resource_id(Text, optional): The user name and policy name with a separator '-'. Format: [user_name]-[policy_name].
         If not specified, Idem will use "name" parameter to identify the IAM policy on AWS.

    Request Syntax:
        [iam-user-policy-name]:
          aws.iam.user_policy.present:
          - name: 'string'
          - resource_id: 'string'
          - user_name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-user-policy:
              aws.iam.user_policy.absent:
                - name: idem-test-user-policy
                - resource_id: idem-test-user-idem-test-user-policy
                - user_name: idem-test-user
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not user_name:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.user_policy", name=name
        )
        return result
    policy_name = name
    if resource_id:
        if not re.search(f"^({user_name})-({policy_name})$", resource_id):
            result[
                "comment"
            ] = f"Incorrect aws.iam.user_policy resource_id: {resource_id}. Expected id {user_name}-{policy_name}"
            result["result"] = False
            return result

    resource = hub.tool.boto3.resource.create(
        ctx,
        "iam",
        "UserPolicy",
        user_name=user_name,
        name=policy_name,
    )
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = (
            f"aws.iam.user_policy '{policy_name}' for user '{user_name}' already absent",
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_policy_to_present(before)

        if ctx.get("test", False):
            result["comment"] = (
                f"Would delete aws.iam.user_policy '{policy_name}' for user '{user_name}'",
            )
            return result

        ret = await hub.exec.boto3.client.iam.delete_user_policy(
            ctx,
            UserName=user_name,
            PolicyName=policy_name,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = (
            f"Deleted aws.iam.user_policy '{policy_name}' for user '{user_name}'",
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the names of the inline policies embedded in the specified IAM user. An IAM user can also have managed
    policies attached to it. To list the managed policies that are attached to a user, use ListAttachedUserPolicies.
    For more information about policies, see Managed policies and inline policies in the IAM User Guide. You can
    paginate the results using the MaxItems and Marker parameters. If there are no inline policies embedded with the
    specified user, the operation returns an empty list.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.iam.user_policy
    """

    result = {}
    # Fetch all users and then policies for each of them
    users = await hub.exec.boto3.client.iam.list_users(ctx)
    if not users["result"]:
        hub.log.debug(f"Could not describe users {users['comment']}")
        return {}

    for user in users["ret"]["Users"]:
        user_name = user.get("UserName")

        # List all user policy names of each user
        try:
            user_policies = await hub.exec.boto3.client.iam.list_user_policies(
                ctx, UserName=user_name
            )
            if not user_policies["result"]:
                hub.log.warning(
                    f"Could not fetch user policies for user {user_name} "
                    f"due to error {user_policies['comment']}. Skip this user and continue."
                )
                continue
        except hub.tool.boto3.exception.ClientError as e:
            hub.log.warning(
                f"Could not fetch user policies for user {user_name} due to error"
                f" {e.__class__.__name__}: {e}. Skip this user and continue."
            )
            continue
        if user_policies:
            if not user_policies["result"]:
                hub.log.warning(
                    f"Could not describe user_policy for user {user_name} due to error "
                    f"{user_policies['comment']}. Skip this user and continue."
                )
            else:
                for user_policy in asyncio.as_completed(
                    [
                        hub.exec.boto3.client.iam.get_user_policy(
                            ctx=ctx, UserName=user_name, PolicyName=user_policy_name
                        )
                        for user_policy_name in user_policies["ret"].get(
                            "PolicyNames", []
                        )
                    ]
                ):
                    ret_user_policy = await user_policy
                    if not ret_user_policy["result"]:
                        hub.log.warning(
                            f"Could not get user_policy for user {user_name} due to error"
                            f" {ret_user_policy['comment']} . Skip this user policy and continue."
                        )
                    else:
                        resource = ret_user_policy["ret"]
                        resource_translated = hub.tool.aws.iam.conversion_utils.convert_raw_user_policy_to_present(
                            resource
                        )
                        resource_key = (
                            f"iam-user-policy-{resource_translated['resource_id']}"
                        )
                        result[resource_key] = {
                            "aws.iam.user_policy.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in resource_translated.items()
                            ]
                        }
    return result


async def _update_user_policy(
    hub, ctx, user_name: str, name: str, policy_document
) -> Dict[str, Any]:
    result = dict(comment=(), result=True)

    ret = await hub.exec.boto3.client.iam.put_user_policy(
        ctx,
        UserName=user_name,
        PolicyName=name,
        PolicyDocument=policy_document,
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]

    return result



import copy
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str = None):
    """
    Get an IAM role from AWS. Provide role name as input

    Args:
        name(string): An Idem name of the IAM role. Idem will use this as the role name if resource_id is not specified.
        resource_id(string, optional): AWS IAM role name.

    """
    result = dict(comment=[], ret=None, result=True)
    before = await hub.exec.boto3.client.iam.get_role(
        ctx, RoleName=resource_id if resource_id else name
    )
    if not before["result"]:
        if "NoSuchEntity" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.role", name=name
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    before_tag = await hub.exec.boto3.client.iam.list_role_tags(
        ctx, RoleName=resource_id if resource_id else name
    )
    if not before_tag["result"]:
        result["result"] = False
        result["comment"] = before_tag["comment"]
        return result
    if before["ret"].get("Role"):
        before["ret"]["Role"]["Tags"] = before_tag["ret"].get("Tags", [])
        result["ret"] = hub.tool.aws.iam.conversion_utils.convert_raw_role_to_present(
            raw_resource=before["ret"]["Role"]
        )
    return result


async def update_role_tags(
    hub,
    ctx,
    role_name: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Update tags of AWS IAM Role

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        role_name: AWS IAM role name
        old_tags: dict of old tags
        new_tags: dict of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict of updated tags}
    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_role(
                ctx, RoleName=role_name, TagKeys=list(tags_to_remove.keys())
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_role(
                ctx,
                RoleName=role_name,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result


async def update_role(
    hub,
    ctx,
    old_state: Dict[str, Any],
    description: str = None,
    max_session_duration: int = None,
    timeout: Dict = None,
):
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    if (description is not None) and old_state.get("description") != description:
        update_payload["Description"] = description
    if (max_session_duration is not None) and old_state.get(
        "max_session_duration"
    ) != max_session_duration:
        update_payload["MaxSessionDuration"] = max_session_duration
    if update_payload:
        if not ctx.get("test", False):
            role_name = old_state.get("name")
            update_ret = await hub.exec.boto3.client.iam.update_role(
                ctx=ctx, RoleName=role_name, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result

            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=1,
                default_max_attempts=40,
                timeout_config=timeout.get("update") if timeout else None,
            )
            hub.log.debug(f"Waiting on updating aws.iam.role '{role_name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "iam",
                    "role_exists",
                    RoleName=role_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
        result["ret"] = {}
        if "MaxSessionDuration" in update_payload:
            result["ret"]["max_session_duration"] = update_payload["MaxSessionDuration"]
            result["comment"] = result["comment"] + (
                f"Update max_session_duration: {update_payload['MaxSessionDuration']}",
            )
        if "Description" in update_payload:
            result["ret"]["description"] = update_payload["Description"]
            result["comment"] = result["comment"] + (
                f"Update description: {update_payload['Description']}",
            )
    return result


async def update_policy(
    hub,
    ctx,
    role_name: str,
    policy: str,
):
    # Update role's trust policy (the policy that is embedded within the role)
    result = dict(comment=(), result=True, ret=None)

    if not ctx.get("test", False):
        update_ret = await hub.exec.boto3.client.iam.update_assume_role_policy(
            ctx=ctx, RoleName=role_name, PolicyDocument=policy
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
            return result

    result["ret"] = policy
    if ctx.get("test", False):
        result["comment"] = (f"Would update aws.iam.role '{role_name}' policy",)
    else:
        result["comment"] = (f"Updated aws.iam.role '{role_name}' policy",)
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions to convert raw resource state from AWS EC2 to present input format.
"""


def convert_raw_db_subnet_group_to_present(
    hub, resource: Dict[str, Any], tags: List = None, idem_resource_name: str = None
) -> Dict[str, Any]:
    new_resource = {}
    db_subnet_group_id = resource.get("DBSubnetGroupName")
    describe_parameters = OrderedDict(
        {
            "DBSubnetGroupDescription": "db_subnet_group_description",
            "DBSubnetGroupArn": "db_subnet_group_arn",
        }
    )
    new_resource = {"name": db_subnet_group_id, "resource_id": db_subnet_group_id}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if resource.get(parameter_old_key) is not None:
            new_resource[parameter_new_key] = resource.get(parameter_old_key)
    if resource.get("Subnets"):
        subnets = []
        for each_subnet in resource.get("Subnets"):
            subnets.append(each_subnet.get("SubnetIdentifier"))

        new_resource["subnets"] = subnets
    if tags:
        new_resource["tags"] = tags

    new_resource["subnets"].sort()
    return new_resource


def convert_raw_db_cluster_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:

    resource_parameters = OrderedDict(
        {
            "AvailabilityZones": "availability_zone",
            "BackupRetentionPeriod": "backup_retention_period",
            "CharacterSetName": "character_set_name",
            "DatabaseName": "database_name",
            "DBClusterIdentifier": "db_cluster_identifier",
            "Engine": "engine",
            "EngineVersion": "engine_version",
            "Port": "port",
            "MasterUsername": "master_username",
            "PreferredBackupWindow": "preferred_backup_window",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "ReplicationSourceIdentifier": "replication_source_identifier",
            "StorageEncrypted": "storage_encrypted",
            "KmsKeyId": "kms_key_id",
            "BacktrackWindow": "backtrack_window",
            "EnableCloudwatchLogsExports": "enable_cloudwatch_logs_exports",
            "EngineMode": "engine_mode",
            "DeletionProtection": "deletion_protection",
            "CopyTagsToSnapshot": "copy_tags_to_snapshot",
            "DBClusterInstanceClass": "db_cluster_instance_class",
            "AllocatedStorage": "allocated_storage",
            "StorageType": "storage_type",
            "Iops": "iops",
            "PubliclyAccessible": "publicly_accessible",
            "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
            "MonitoringInterval": "monitoring_interval",
            "MonitoringRoleArn": "monitoring_role_arn",
            "PerformanceInsightsKMSKeyId": "performance_insights_kms_key_id",
            "PerformanceInsightsRetentionPeriod": "performance_insights_retention_period",
            "DBClusterParameterGroup": "db_cluster_parameter_group_name",
            "DBSubnetGroup": "db_subnet_group_name",
            "TagList": "tags",
            "ScalingConfigurationInfo": "scaling_configuration",
        }
    )
    resource_id = raw_resource.get("DBClusterIdentifier")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("VpcSecurityGroups"):
        vpc_security_group_id = []
        for vpc_security_group in raw_resource.get("VpcSecurityGroups"):
            if "VpcSecurityGroupId" in vpc_security_group:
                vpc_security_group_id.append(
                    vpc_security_group.get("VpcSecurityGroupId")
                )
        resource_translated["vpc_security_group_ids"] = vpc_security_group_id

    if raw_resource.get("DBClusterOptionGroupMemberships"):
        for option_group_membership in raw_resource.get(
            "DBClusterOptionGroupMemberships"
        ):
            resource_translated["option_group_name"] = option_group_membership.get(
                "DBClusterOptionGroupName"
            )
            break

    if raw_resource.get("DomainMemberships"):
        for domain_membership in raw_resource.get("DomainMemberships"):
            resource_translated["domain"] = domain_membership.get("Domain")
            resource_translated["domain_iam_role_name"] = domain_membership.get(
                "IAMRoleName"
            )
            break

    if resource_translated["tags"]:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            resource_translated["tags"]
        )
    return resource_translated


async def convert_raw_db_instance_to_present(
    hub,
    raw_resource: Dict[str, Any],
    raw_resource_tags: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:

    resource_parameters = OrderedDict(
        {
            "DBName": "db_name",
            "AllocatedStorage": "allocated_storage",
            "DBInstanceClass": "db_instance_class",
            "Engine": "engine",
            "MasterUsername": "master_username",
            "AvailabilityZone": "availability_zone",
            "PreferredMaintenanceWindow": "preferred_maintenance_window",
            "BackupRetentionPeriod": "backup_retention_period",
            "PreferredBackupWindow": "preferred_backup_window",
            "MultiAZ": "multi_az",
            "EngineVersion": "engine_version",
            "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
            "LicenseModel": "license_model",
            "Iops": "iops",
            "CharacterSetName": "character_set_name",
            "NcharCharacterSetName": "nchar_character_set_name",
            "PubliclyAccessible": "publicly_accessible",
            "DBClusterIdentifier": "db_cluster_identifier",
            "StorageType": "storage_type",
            "StorageEncrypted": "storage_encrypted",
            "KmsKeyId": "kms_key_id",
            "CopyTagsToSnapshot": "copy_tags_to_snapshot",
            "MonitoringInterval": "monitoring_interval",
            "MonitoringRoleArn": "monitoring_role_arn",
            "PromotionTier": "promotion_tier",
            "Timezone": "timezone",
            "PerformanceInsightsKMSKeyId": "performance_insights_kms_key_id",
            "PerformanceInsightsRetentionPeriod": "performance_insights_retention_period",
            "EnableCloudwatchLogsExports": "enable_cloudwatch_logs_exports",
            "ProcessorFeatures": "processor_features",
            "DeletionProtection": "deletion_protection",
            "MaxAllocatedStorage": "max_allocated_storage",
            "CustomIamInstanceProfile": "custom_iam_instance_profile",
            "BackupTarget": "backup_target",
        }
    )
    resource_id = raw_resource.get("DBInstanceIdentifier")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("DBSecurityGroups"):
        db_security_group_names = []
        for db_security_group in raw_resource.get("DBSecurityGroups"):
            if "DBSecurityGroupName" in db_security_group:
                db_security_group_names.append(
                    db_security_group.get("DBSecurityGroupName")
                )
        resource_translated["db_security_groups"] = db_security_group_names

    if raw_resource.get("VpcSecurityGroups"):
        vpc_security_group_id = []
        for vpc_security_group in raw_resource.get("VpcSecurityGroups"):
            if "VpcSecurityGroupId" in vpc_security_group:
                vpc_security_group_id.append(
                    vpc_security_group.get("VpcSecurityGroupId")
                )
        resource_translated["vpc_security_group_ids"] = vpc_security_group_id

    if raw_resource.get("DBSubnetGroup"):
        db_subnet_group = raw_resource.get("DBSubnetGroup")
        if "DBSubnetGroupName" in db_subnet_group:
            resource_translated["db_subnet_group_name"] = db_subnet_group.get(
                "DBSubnetGroupName"
            )

    if raw_resource.get("DBParameterGroups"):
        for db_parameter_group in raw_resource.get("DBParameterGroups"):
            resource_translated["db_parameter_group_name"] = db_parameter_group.get(
                "DBParameterGroupName"
            )
            break

    if raw_resource.get("Endpoint"):
        resource_translated["port"] = raw_resource.get("Endpoint").get("Port")

    if raw_resource.get("OptionGroupMemberships"):
        for option_group_membership in raw_resource.get("OptionGroupMemberships"):
            resource_translated["option_group_name"] = option_group_membership.get(
                "OptionGroupName"
            )
            break

    if raw_resource.get("DomainMemberships"):
        for domain_memberships in raw_resource.get("DomainMemberships"):
            resource_translated["domain"] = domain_memberships.get("Domain")
            resource_translated["domain_iam_role_name"] = domain_memberships.get(
                "IAMRoleName"
            )
            break

    if raw_resource_tags.get("ret") and raw_resource_tags.get("ret").get("TagList"):
        resource_translated["tags"] = raw_resource_tags.get("ret").get("TagList")

    if resource_translated.get("tags") is not None:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            resource_translated.get("tags")
        )

    return resource_translated


async def convert_db_parameter_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    tags: List = None,
    parameters: List = None,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    new_resource = {}
    db_parameter_group_id = idem_resource_name
    describe_parameters = OrderedDict(
        {
            "DBParameterGroupName": "name",
            "DBParameterGroupFamily": "db_parameter_group_family",
            "Description": "description",
            "DBParameterGroupArn": "db_parameter_group_arn",
        }
    )
    new_resource = {"name": db_parameter_group_id, "resource_id": db_parameter_group_id}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            new_resource[parameter_new_key] = raw_resource.get(parameter_old_key)
    if tags:
        new_resource["tags"] = tags
    if parameters:
        db_parameters = OrderedDict(
            {"ParameterName": "name", "ParameterValue": "value"}
        )
        list_parameters = []
        for parameter in parameters:
            update_parameters = {}
            for parameter_old_key, parameter_new_key in db_parameters.items():
                if parameter.get(parameter_old_key) is not None:
                    update_parameters[parameter_new_key] = parameter.get(
                        parameter_old_key
                    )
            list_parameters.append(update_parameters)
        new_resource["parameters"] = list_parameters
    return new_resource


"""
Util function to convert raw resource state from AWS RDS to present input format.
"""


def convert_raw_db_cluster_parameter_group_to_present(
    hub, raw_resource: Dict[str, Any], tags: List = None
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "DBClusterParameterGroupName": "resource_id",
            "DBParameterGroupFamily": "db_parameter_group_family",
            "Description": "description",
        }
    )
    resource_translated = {"name": raw_resource.get("DBClusterParameterGroupName")}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if tags:
        resource_translated["tags"] = tags

    return resource_translated
